#!/bin/sh

format_ext() {
	local dev="$1"
	local label="$2"

	[ -e "$dev" ] || error "No partition $dev!"
	dd if=/dev/zero of="$dev" bs=1M count=1
	mkfs.ext4 -L "$label" "$dev" \
		|| error "mkfs ext4 failed"
}

copy_firm() {
	local dev="$1"

	if [ -e "/mnt/firm.squashfs" ]; then
		[ -e "$dev" ] || error "No partition $dev!"
		dd if=/mnt/firm.squashfs of="$dev" bs=1M \
			|| error "Copy of firmware failed"
	fi
	if unshare -m mount "$dev" /mnt >/dev/null 2>&1; then
		firmware_present=1
	fi
}

format_app() {
	local dev="$1"
	local label="app"

	[ -e "$dev" ] || error "No partition $dev!"
	dd if=/dev/zero of="$dev" bs=1M count=1
	mkfs.btrfs -L "$label" -m DUP -R free-space-tree "$dev" \
		|| error "mkfs btrfs failed"
	mount "$dev" /mnt
	btrfs subvolume create /mnt/tmp \
		|| error "subvolume creation failed"
	mkdir /mnt/boot_0 /mnt/boot_1
	btrfs subvolume create /mnt/volumes \
		|| error "subvolume creation failed"
	btrfs subvolume create /mnt/boot_0/containers_storage \
		|| error "subvolume creation failed"
	btrfs subvolume create /mnt/boot_0/volumes \
		|| error "subvolume creation failed"

	# work around podman being silly and throwing errors if the
	# store does not look right
	podman run --storage-opt additionalimagestore="" \
		--root /mnt/boot_0/containers_storage image list >/dev/null 2>&1
	rm -f /mnt/boot_0/containers_storage/libpod/bolt_state.db
	# yes it really needs that too...
	mkdir -p /mnt/boot_0/containers_storage/overlay-layers /mnt/boot_0/containers_storage/overlay-images
	touch /mnt/boot_0/containers_storage/overlay-images/images.lock
	touch /mnt/boot_0/containers_storage/overlay-layers/layers.lock
	btrfs property set /mnt/boot_0/containers_storage ro true

	# and prepare boot_1 as well just in case
	btrfs subvolume snapshot -r /mnt/boot_0/containers_storage /mnt/boot_1/containers_storage \
		|| error "subvolume creation failed"
	btrfs subvolume create /mnt/boot_1/volumes \
		|| error "subvolume creation failed"

	mountpoint -q /mnt/boot_0/containers_storage/overlay \
		&& umount /mnt/boot_0/containers_storage/overlay
	umount /mnt
}

create_partitions() {
	local part start size

	# We need a first non-trivial sgdisk call to "resize" the gpt,
	# otherwise partition creations fail saying there is not
	# enough space... Rerandomize disk/partition GUID.
	sgdisk -G "$DISK"

	size="$(lsblk -o SIZE -rn "$DISKPART"1)"
	sgdisk --new 2:0:+"$size" -c 2:rootfs_1 \
		--new 3:0:+50M -c 3:logs \
		--new 4:0:+200M -c 4:firm \
		--new 5:0:0 -c 5:app \
		"$DISK" \
		|| error "partitionning disk failed"

	# first partition is in use, use losetup -P to
	# recreate a new device with its own partition table
	loop=$(losetup --show -f -P "$DISK")
	format_ext "${loop}p3" logs
	copy_firm "${loop}p4"
	format_app "${loop}p5"
	losetup -d "$loop"
	loop=""

	# format gp partitions we use
	if [ -e "$DISK"gp0 ]; then
		mkfs.ext4 -F "$DISK"gp0
		# XXX copy licenses
	fi
	if [ -e "$DISK"gp1 ]; then
		mkfs.fat --mbr=n -n ATLOG "$DISK"gp1
	fi
}

ext_board_eeprom_read() {
	i2cget -y 3 0x50 "$1" b
}

ext_board_is_atmark_techno() {
	# check vendor id is 0x0001 on big endian
	[ "$(ext_board_eeprom_read 0x00)" = 0x00 ] \
	&& [ "$(ext_board_eeprom_read 0x01)" = 0x01 ]
}

ext_board_is_lte() {
	# vendor id must be atmark techno and
	# product id is 0x0001 on big endian
	ext_board_is_atmark_techno \
	&& [ "$(ext_board_eeprom_read 0x02)" = 0x00 ] \
	&& [ "$(ext_board_eeprom_read 0x03)" = 0x01 ]
}

# required environment variables
#  DISK = base block device e.g. /dev/mmcblk2 or /dev/vda
#  DISKPART = prefix for partitions e.g. /dev/mmcblk2p or /dev/vda
firstboot() {
	local boot boot_version firmware_present=""
	[ "${DISK#/dev/mmcblk}" != "$DISK" ] && DISK="${DISK%p[0-9]}" || DISK="$DISKPART"

	create_partitions

	# self-destruct, and poweroff only if we really did something
	mount "${DISKPART}1" /mnt
	mount -o remount,rw /mnt
	case "$DISK" in
	/dev/mmcblk2)
		boot="cat /dev/mmcblk2boot0"
		if [ -e /dev/mmcblk2gp1 ]; then
			cat >> /mnt/etc/fstab <<'EOF'
/dev/mmcblk2gp1	/var/at-log			vfat	defaults			0 0
EOF
		fi
		;;
	/dev/mmcblk1)
		boot="dd if=/dev/mmcblk1 bs=1k skip=32 count=4096"
		sed -i -e "s:/dev/mmcblk2p:$DISKPART:" /mnt/etc/fstab
		sed -i -e "s:/dev/mmcblk2boot0:$DISK:" /mnt/etc/fw_env.config
		;;
	*)
		# no env for scsi/sata disks
		sed -i -e 's:/dev/mmcblk2p3:LABEL=logs:' -e 's:/dev/mmcblk2p5:LABEL=app:' /mnt/etc/fstab
		> /mnt/etc/fw_env.config
		rm -f /mnt/etc/runlevels/default/reset_bootcount
		;;
	esac

	# update versions
	[ -e "/mnt/etc/atmark-release" ] \
		&& echo "base_os $(cat "/mnt/etc/atmark-release")" >> /mnt/etc/sw-versions
	if [ -n "$boot" ]; then
		boot_version=$(eval "$boot" | strings |
			grep -m1 -oE '20[0-9]{2}.[0-1][0-9]-[0-9a-zA-Z.-]*')
		[ -n "$boot_version" ] \
			&& echo "boot $boot_version" >> /mnt/etc/sw-versions
	fi

	if grep -q uninit /mnt/etc/hwrevision; then
		local board
		board=$(dd if=/sys/bus/nvmem/devices/imx-ocotp0/nvmem bs=4 skip=56 count=1 status=none | xxd -p | head -c 4)
		case "$board" in
		c600) echo 'iot-g4-eva at1' > /mnt/etc/hwrevision;;
		c700) echo 'iot-g4-es1 at1' > /mnt/etc/hwrevision;;
		c800) echo 'iot-g4-es2 at1' > /mnt/etc/hwrevision;;
		c900) echo 'AGX4500 at1' > /mnt/etc/hwrevision;;
		cb00) echo 'iot-g4-es3 at1' > /mnt/etc/hwrevision;;
		esac
	fi

	if [ -n "$firmware_present" ]; then
		cat >> /mnt/etc/fstab <<EOF
${DISKPART}4	/opt/firmware			squashfs defaults			0 0
EOF
		rm -f /mnt/firm.squashfs
	fi

	if ext_board_is_lte; then
		einfo "LTE extension board exists"
		echo "fdt_overlays=armadillo_iotg_g4-lte-ext-board.dtbo" > /mnt/boot/overlays.txt
		ln -s /etc/init.d/connection-recover /mnt/etc/runlevels/default/
		ln -s /etc/init.d/modemmanager /mnt/etc/runlevels/boot/
	fi

	rm -f /mnt/etc/init.d/firstboot-atmark
	rm -f /mnt/etc/runlevels/sysinit/firstboot-atmark
	rm -f /mnt/lib/rc/sh/functions-atmark.sh
	date +%s.%N > /mnt/etc/.rootfs_update_timestamp
	umount /mnt
}
